/*--------------------------------------------------------------------------+
$Id: codetemplates.xml 1 2011-01-01 00:00:01Z hoelzl $
|                                                                          |
| Copyright 2011 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.decompose;

import java.util.ArrayList;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.fortiss.tooling.base.model.element.IConnection;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHiddenSpecification;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.element.IHierarchicElementContainer;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.model.element.IModelElementSpecification;
import org.fortiss.tooling.kernel.extension.IElementCompositor;
import org.fortiss.tooling.kernel.service.IConnectionCompositorService;
import org.fortiss.tooling.kernel.service.IElementCompositorService;

/**
 * Base implementation of a compositor for {@link IHierarchicElementContainer}s
 * focusing on the remove mechanism.
 * 
 * @author doebber
 * @author $Author: hoelzl $
 * @version $Rev: 18709 $
 * @ConQAT.Rating RED Hash:
 */
public abstract class HierarchicElementCompositorBase<HE extends IHierarchicElementContainer>
		implements IElementCompositor<HE> {

	/**
	 * Decomposition ability check that takes care of the whole model subtree.
	 * Subclasses implement {@link #canDecompose*****()} instead.
	 */
	@Override
	public final boolean canDecompose(EObject contained) {
		// TODO remove
		System.out.println(contained);
		return iterateDecompose(contained, true);
	}

	/**
	 * Decomposition ability check that takes care of the whole model subtree.
	 * Subclasses implement {@link #decompose*****()} instead.
	 */
	@Override
	public final boolean decompose(EObject contained) {
		return iterateDecompose(contained, false);
	}

	/**
	 * Returns whether the compositor is able to decompose the given model
	 * element
	 */
	protected abstract boolean canDecomposeSpecific(EObject contained);

	/**
	 * Base implementation for element removal using EcoreUtil.delete
	 * (non-recursively). Subclasses may override.
	 */
	protected boolean decomposeSpecific(EObject contained) {
		System.out.println("specific: " + contained + " ### " + this);
		EcoreUtil.delete(contained);
		return true;
	}

	/** recursive traversion of all model elements in subtree */
	private boolean iterateDecompose(EObject contained, boolean can) {

		if (!canDecomposeSpecific(contained)) {
			return false;
		}

		if (can) {
			if (contained instanceof IModelElement) {
				if (!iterateCanDecomposeSpecifications((IModelElement) contained)) {
					return false;
				}
				// TODO CD: handle references
			}
			if (contained instanceof IHierarchicElementContainer) {
				if (!iterateCanDecomposeSubelements((IHierarchicElementContainer) contained)) {
					return false;
				}
			}
			if (contained instanceof IHierarchicElement) {
				IHierarchicElement he = (IHierarchicElement) contained;

				if (!iterateCanDecomposeConnections(he.getConnectionsList())) {
					return false;
				}

				if (!iterateCanDecomposeConnectors((IHierarchicElement) contained)) {
					return false;
				}

			}
			if (contained instanceof IConnector) {
				IConnector connector = (IConnector) contained;
				if (!iterateCanDecomposeConnections(connector.getIncomingList())) {
					return false;
				}
				if (!iterateCanDecomposeConnections(connector.getOutgoingList())) {
					return false;
				}
			}
		} else {
			if (contained instanceof IModelElement) {
				if (!iterateDecomposeSpecifications((IModelElement) contained)) {
					return false;
				}
				// TODO CD: handle references
			}
			if (contained instanceof IHierarchicElementContainer) {
				if (!iterateDecomposeSubelements((IHierarchicElementContainer) contained)) {
					return false;
				}
			}
			if (contained instanceof IHierarchicElement) {
				IHierarchicElement he = (IHierarchicElement) contained;

				if (!iterateDecomposeConnections(he.getConnectionsList())) {
					return false;
				}

				if (!iterateDecomposeConnectors((IHierarchicElement) contained)) {
					return false;
				}

			}
			if (contained instanceof IConnector) {
				IConnector connector = (IConnector) contained;
				if (!iterateDecomposeConnections(connector.getIncomingList())) {
					return false;
				}
				if (!iterateDecomposeConnections(connector.getOutgoingList())) {
					return false;
				}
			}
			return decomposeSpecific(contained);
		}

		return true;
	}

	/**
	 * Iterates over specifications of given {@link IModelElement} and checks
	 * for decomposition ability
	 */
	private boolean iterateCanDecomposeSpecifications(IModelElement me) {
		for (IModelElementSpecification spec : me.getSpecificationsList()) {
			if (!canDecomposeSpecification(spec)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Iterates over sub elements of given {@link IHierarchicElementContainer}
	 * and checks for decomposition ability.
	 */
	private boolean iterateCanDecomposeSubelements(
			IHierarchicElementContainer contained) {
		for (IHierarchicElement subelement : contained
				.getContainedElementsList()) {
			if (!canDecomposeSubElement(subelement)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Iterates over list of connections and checks for disconnection ability or
	 * decomposition ability of sub elements respectively.
	 */
	private boolean iterateCanDecomposeConnections(
			EList<IConnection> connectionsList) {
		for (IConnection conn : connectionsList) {
			if (!canDecomposeConnection(conn)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Iterates over connectors of given {@link IHierarchicElement} and checks
	 * for decomposition ability.
	 */
	private boolean iterateCanDecomposeConnectors(IHierarchicElement contained) {
		for (IConnector c : contained.getConnectorsList()) {
			if (!canDecomposeConnector(c)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Iterates over specifications of given {@link IModelElement} and checks
	 * for decomposition ability
	 */
	private boolean iterateDecomposeSpecifications(IModelElement me) {
		ArrayList<IModelElementSpecification> list = new ArrayList<IModelElementSpecification>(
				me.getSpecificationsList());
		for (IModelElementSpecification spec : list) {
			if (!decomposeSpecification(spec)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Iterates over sub elements of given {@link IHierarchicElementContainer}
	 * and checks for decomposition ability.
	 */
	private boolean iterateDecomposeSubelements(
			IHierarchicElementContainer contained) {
		ArrayList<IHierarchicElement> list = new ArrayList<IHierarchicElement>(
				contained.getContainedElementsList());
		for (IHierarchicElement subelement : list) {
			if (!decomposeSubelement(subelement)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Iterates over list of connections and checks for disconnection ability or
	 * decomposition ability of sub elements respectively.
	 */
	private boolean iterateDecomposeConnections(
			EList<IConnection> connectionsList) {
		ArrayList<IConnection> list = new ArrayList<IConnection>(
				connectionsList);
		for (IConnection conn : list) {
			if (!decomposeConnection(conn)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Iterates over connectors of given {@link IHierarchicElement} and checks
	 * for decomposition ability.
	 */
	private boolean iterateDecomposeConnectors(IHierarchicElement contained) {
		ArrayList<IConnector> list = new ArrayList<IConnector>(
				contained.getConnectorsList());
		for (IConnector c : list) {
			if (!decomposeConnector(c)) {
				return false;
			}
		}
		return true;
	}

	/** Base implementation returns true by default. Subclasses may override. */
	protected boolean canDecomposeSpecification(
			IModelElementSpecification element) {
		if (element instanceof IHiddenSpecification) {
			return true;
		}
		return IElementCompositorService.INSTANCE.canDecompose(element);
	}

	/** Base implementation returns true by default. Subclasses may override. */
	protected boolean canDecomposeReference(EObject element) {
		// TODO implement reference removal
		return true;
	}

	/** Base implementation returns true by default. Subclasses may override. */
	protected boolean canDecomposeSubElement(IHierarchicElement element) {
		return IElementCompositorService.INSTANCE.canDecompose(element);
	}

	/** Base implementation returns true by default. Subclasses may override. */
	protected boolean canDecomposeConnector(IConnector element) {
		return IElementCompositorService.INSTANCE.canDecompose(element);
	}

	/**
	 * Base implementation returns compositors' decision by default. Subclasses
	 * may override.
	 */
	protected boolean canDecomposeConnection(IConnection conn) {
		return IConnectionCompositorService.INSTANCE.canDisconnect(conn);
	}

	protected boolean decomposeSpecification(IModelElementSpecification element) {
		System.out.println("specification: " + element + " ### " + this);
		EcoreUtil.delete(element);
		return true;
	}

	protected boolean decomposeReferences(EObject element) {
		// TODO implement reference removal
		return true;
	}

	protected boolean decomposeSubelement(IHierarchicElement element) {
		System.out.println("subelement: " + element + " ### " + this);
		if (IElementCompositorService.INSTANCE.canDecompose(element)) {
			return IElementCompositorService.INSTANCE.decompose(element);
		}
		EcoreUtil.delete(element);
		return true;
	}

	protected boolean decomposeConnector(IConnector element) {
		ArrayList<IConnection> list = new ArrayList<IConnection>(
				element.getIncomingList());
		list.addAll(element.getOutgoingList());
		for (IConnection conn : list) {
			System.out.println("disconnect: " + element + " ### " + this);
			if (!IConnectionCompositorService.INSTANCE.disconnect(conn)) {
				return false;
			}
		}
		EcoreUtil.delete(element);
		return true;
	}

	protected boolean decomposeConnection(IConnection conn) {
		System.out.println("connection: " + conn + " ### " + this);
		return IConnectionCompositorService.INSTANCE.disconnect(conn);
	}

}
